#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#define NUM_FILES 12                         /* max # of devices to test */
#define OFFSET 262144
#define ALARM_TIME 30                        /* length of testing (seconds) */

key_t cnt_key,init_key;                      /* shared memory queue keys    */
                                             /* shared memory segment IDs   */
int shmcnt,                                  /* for IO counter              */
    initcnt;                                 /* for child processes started */
                                             /* pointers to shared memory   */
int *io_cnt,                                 /* stores # of IOs completed   */
    *init_cnt;                               /* stores # of child processes */

char *buf;                                   /* buffer used for I/O */
int buf_sz,                                  /* size of I/Os to be issued */
    devnum;                                  /* device process will exercise */

int files_under_test[NUM_FILES],             /* array of file names to test */
    num_files,                               /* number of devices to test */
    num_blks,                                /* seek range in I/O sized blks */
    num_children,                            /* number of processes doing IO */
    read_per,                                /* percent of IOs that are reads */
    finish_flag=0,                           /* test until flag is set */
    stored_ios[11],                          /* array of I/Os completed/pass */
    num_reps=0;                              /* number of passes required to */
                                             /* get a consistent result I/O  */
long seed;                                   /* random number seed */

/*******************************************************************/
sigusr1_handler()
{
signal(SIGUSR1,sigusr1_handler);
return;
}
/*******************************************************************/
/*******************************************************************/
child_sigusr2_handler()
{
shmdt(io_cnt);                               /* detach shared memory variable */
exit(0);
}
/*******************************************************************/
/*******************************************************************/
parent_alarm_handler()
{
int steady_val;
int diff;
static int begin_num_ios;
static int end_num_ios;

end_num_ios=io_cnt[1];                       /* IOs  completed on last pass */
stored_ios[num_reps]=end_num_ios-begin_num_ios;  /* keep IO pass totals */
steady_val=stored_ios[num_reps] / 100;       /* get 1% of last IO pass total */
if(num_reps >= 1)                            /* have we done this one time?  */
   {                                         /* get the difference between   */
   diff=stored_ios[num_reps] - stored_ios[num_reps-1]; /* the last pass and  */
   diff=abs(diff);                           /* the current pass             */
   if((diff <= steady_val) || (num_reps >=10)) /* is the diff between 1% or  */
      {                                      /* have we tried 10 times       */
      finish_flag=1;                         /* if yes, set flag to done     */
      kill(0,SIGUSR2);                       /* send children kill signal in */
      return;                                /* while not finished loop      */
      }
   }
num_reps++;                                  /* increment number of passes */
signal(SIGALRM,parent_alarm_handler);        /* get setup for next alarm */
begin_num_ios=io_cnt[1];                     /* store total IOs completed */
}
/*******************************************************************/

main(argc,argv)
int argc;
char *argv[];
{
int num_ios,i=0,wait_stat,tmp;

struct shmid_ds shmid_ds_buf;


float ios_sec,
      resp_time,
      throughput,
      tot_time,
      flt_num_ios;

set_up(argc,argv);                           /* get cmd line & init shar mem */

num_children=atoi(argv[1]);                  /* how many IO processes to run */

start_children(argv,files_under_test);       /* start child IO processes */

while(init_cnt[1] < num_children)            /* wait here until all of the   */
   {                                         /* child processes have started */
   sleep(1);
   }

signal(SIGUSR2,SIG_IGN);                     /* parent will ignore user sig 2 */
signal(SIGALRM,parent_alarm_handler);        /* do totals when ALARM rings */
kill(0,SIGUSR1);                             /* signal children to start */

while(finish_flag != 1)                      /* loop here until consistent    */
   {                                         /* results from child processes  */
   alarm(ALARM_TIME);                        /* ring alarm in num of seconds  */
   pause();                                  /* do nothing until ALARM rings  */
   alarm(0);                                 /* cancel previous ALARM request */
   } 

tot_time=ALARM_TIME;                         /* how long did we run for? */
flt_num_ios=stored_ios[num_reps];            /* move last IO pass # to float */
ios_sec=flt_num_ios/tot_time;                /* calculate IOs per second */
resp_time=(1000*tot_time)/flt_num_ios;       
throughput=ios_sec*(float)(buf_sz/1024);     /* calculate KB/sec thruput */
                                             /* print results used by shell */
printf("%d %d 0 0 0 0 0 %.3f %.3f %.3f\n",num_children,num_reps,resp_time,ios_sec,throughput); 

shmdt(io_cnt);                               /* detach shared memory variable */
shmctl(shmcnt,IPC_RMID,&shmid_ds_buf);       /* remove shared memory ID */
shmdt(init_cnt);
shmctl(initcnt,IPC_RMID,&shmid_ds_buf);
}

/*******************************************************************/
set_up(argc,argv)
int argc;
char *argv[];
{
int i;

if(argc < 7)                                 /* check for enough arguments */
   {                                         /* if not, print usage stmt   */
   printf("usage: unixio <numprocs> <read_per> <blksz> <num_blks> <iterations> <file> [files]\n");
   exit(1);
   }
read_per=atoi(argv[2]);                      /* get IO read percentage */
num_blks=atoi(argv[4]);                      /* get address range to test */
num_files=argc-6;                            /* how many files were passed */
buf_sz=atoi(argv[3]);                        /* get IO size to use in test */

for(i = 0; i < num_files; i++)               /* check to see if all of the   */
   {                                         /* files can be written to and  */
   if((files_under_test[i]=open(argv[(6+i)],O_CREAT | O_RDWR)) == -1)
      {                                      /* load cmd line names in array */
      printf("error opening file %s, exiting\n",argv[(6+i)]); /* if problem  */
      exit(2);                               /* opening files, print error   */
      }
   }

/****** Initialize shared memory variables ******/

if((shmcnt=shmget(cnt_key,(2*(sizeof(int))),IPC_CREAT)) == -1)
   {                                         /* create a new region of shared */
   printf("shmget failed,exiting\n");        /* memory and return the seg id  */
   exit(0);
   }
if((io_cnt=shmat(shmcnt,(void *)0,0)) == -1) /* get pointer to sh mem segment */
   {
   printf("shmat failed, exiting\n");        
   exit(0);
   }
io_cnt[1]=0;                                 /* initialize IO completed count */

if((initcnt=shmget(init_key,(2*(sizeof(int))),IPC_CREAT)) == -1)
   {
   printf("shmget failed,exiting\n");
   exit(0);
   }
if((init_cnt=shmat(initcnt,(void *)0,0)) == -1)
   {
   printf("shmat failed, exiting\n");
   exit(0);
   }
init_cnt[1]=0;

if((buf=malloc(buf_sz*(sizeof(char)))) == NULL) /* allocate mem for IO buffer */
   {
   printf("malloc failed,exiting\n");
   exit(0);
   }
for(i = 0; i < buf_sz; i++)               /* fill buffer with something */
   {
   buf[i]='a';
   }
signal(SIGUSR1,sigusr1_handler);          /* initialize user 1 signal routine */
}

/*******************************************************************/
int start_children(argv,files_under_test)
char *argv[];
int *files_under_test;
{
int i=0,
    pid;

devnum=0;

while(i<num_children)                        /* start # of IO processes */
   {                                         /* requested on cmd line */
   if((pid=fork())==-1)                      /* start new copy of this pgm */
      {                                      /* ERROR, could not start */
      printf("fork failed, exiting\n");      /* another instance       */
      exit(0);
      }
   else                                      /* child started OK */
      {
      if(pid == 0)                           /* on fork, child pid value is 0 */
         {                                   /* child process goes this way   */
         signal(SIGUSR2,child_sigusr2_handler);
         init_cnt[1]++;                      /* increment child init counter */
         pause();                            /* wait here until signal */
                                             /* parent will send SIGUSR1 via */
                                             /* kill, children will proceed  */

         seed=getpid();                      /* get unique # for rand # gen */

         do_io(argv,files_under_test);       /* go do I/O */

         shmdt(io_cnt);                      /* detach shared memory variable */
         exit(0);
         }
      else                                   /* parent pid is non-zero */
         {
         i++;                                /* increment loop count for num */
                                             /* of child processes started   */
         devnum = i % num_files;             /* assign device to process     */
         }
      } /* end of fork pid check */
   } /* end of while child loop */
}
         
/*******************************************************************/

int do_io(argv,files_under_test)
char *argv[];
int *files_under_test;
{
int j,read_ret_val,write_ret_val,read_cnt=0,write_cnt=0,extra_seek;
long lseek_place,read_chance;

srand48(seed);                               /* init random num generator */

if(buf_sz >= OFFSET)                         /* still not sure what this is */
   {
   extra_seek=1;
   }
else
   {
   extra_seek=OFFSET/buf_sz;
   }

lseek_place = (lrand48() % num_blks) * buf_sz;

while(1)                                     /* loop forever */
   {
// for(j = 0; j < num_files; j++)            /* loop through the file names */
//    {                                      /* passed on the command line  */
//    lseek_place=lrand48();                 /* get a random number */
      read_chance=lrand48()%100;             /* from a random number, do a   */
                                             /* mod 100 to get a random num  */
                                             /* between 1 and 100            */
//    lseek_place=(lseek_place%num_blks)+extra_seek; /* from the random num, */
                                             /* get a num between 0 & num of */
                                             /* blocks. not sure about xseek */
//    lseek_place*=buf_sz;                   /* addr is multiple of IO size  */
      lseek(files_under_test[devnum],lseek_place,0); /* seek to byte offset */
      if(read_chance < read_per)             /* if rand num is < read_per */
         {                                   /* do a read */
         if((read_ret_val=read(files_under_test[devnum],buf,buf_sz)) != buf_sz)
            {
            printf("read for file %s failed, exiting\n",argv[6+devnum]);
            exit(3);
            }
         else
            {
            io_cnt[1]++;                     /* increment IO complete counter */
            }
         }
      else                                   /* rand num not < read_per */
         {                                   /* do a write */
         if((write_ret_val=write(files_under_test[devnum],buf,buf_sz)) != buf_sz)
            {
            printf("write for file %s failed, exiting\n",argv[6+devnum]);
            exit(4);
            }
         else
            {
            io_cnt[1]++;                     /* increment IO complete counter */
            }
         } /* end IF read_per */
//    } /* end FOR num_files loop */

   if (lseek_place < ((num_blks - 1) * buf_sz)) /* if not end of seek range */
     lseek_place += buf_sz;                     /* go to next seq. address  */
   else
     lseek_place = buf_sz;                      /* else go back to beginning*/
   } /* end WHILE loop */
}
/*******************************************************************/
